from typing import BinaryIO
from struct import unpack

# https://github.com/OpenL2D/moc3ingbird/blob/master/src/moc3.hexpat
def read_moc3(file: BinaryIO):
    '''Reads a MOC3 file and returns Part names and Parameter names

    Args:
        file (BinaryIO): File pointer

    Returns:
        tuple: (list[str], list[str]), Part names and Parameter names
    '''
    # Header: 64 bytes
    file.seek(0)
    assert file.read(4) == b'MOC3'
    version = unpack('<c',file.read(1))[0]
    isBigEndian = unpack('<b',file.read(1))[0]
    assert not isBigEndian
            
    # TODO: Other fields
    file.seek(0x40)
    pCountInfo = unpack('<I',file.read(4))[0]
    
    file.seek(pCountInfo)
    numParts = unpack('<I',file.read(4))[0]
    file.seek(0x10, 1)
    numParameters = unpack('<I',file.read(4))[0]

    file.seek(0x4C)
    pParts = unpack('<I',file.read(4))[0]

    file.seek(0x108)
    pParameters = unpack('<I',file.read(4))[0]
    
    def read_strings(offset, count):
        for i in range(0,count):
            file.seek(offset + i * 0x40)   
            buffer = bytearray()  
            while b := file.read(1)[0]:
                buffer.append(b)
            yield buffer.decode(encoding='utf-8')
    
    return list(read_strings(pParts,numParts)), list(read_strings(pParameters,numParameters))



from os import path,makedirs
from json import loads,dump
from io import BytesIO
from logging import getLogger
from concurrent.futures import ProcessPoolExecutor
from time import sleep
from sssekai.unity.AssetBundle import load_assetbundle
from UnityPy.enums import ClassIDType
from sssekai.fmt.rla import read_rla
from tqdm import tqdm
logger = getLogger(__name__)


def worker_job(sname, version, script):
    rla = read_rla(BytesIO(script), version)
    dump(rla, open(sname + '.json', 'w', encoding='utf-8'), indent=4, ensure_ascii=False)

def main_rla2json(args):
    with open(args.infile,'rb') as f:
        env = load_assetbundle(f)
        datas = dict()
        for obj in env.objects:
            if obj.type in {ClassIDType.TextAsset}:
                data = obj.read()
                datas[data.name] = data 
        header = datas.get('sekai.rlh', None)
        assert header, "RLH Header file not found!"
        makedirs(args.outdir, exist_ok=True)
        header = loads(header.text)
        dump(header, open(path.join(args.outdir, 'sekai.rlh.json'), 'w', encoding='utf-8'), indent=4, ensure_ascii=False)
        version = tuple(map(int, header['version'].split('.')))
        splitSeconds = header['splitSeconds']
        logger.info('Version: %d.%d' % version)
        logger.info('Count: %d' % len(header['splitFileIds']))        
        with ProcessPoolExecutor() as executor:
            logger.info('Dumping RLA data with %d processors' % executor._max_workers)
            futures = []
            for sid in header['splitFileIds']:
                sname = 'sekai_%2d_%08d' % (splitSeconds, sid)
                script = datas[sname + '.rla'].script.tobytes()                
                futures.append(executor.submit(worker_job, path.join(args.outdir,sname), version, script))
            finsihed_futures = set()
            with tqdm(total=len(futures)) as pbar:
                while len(finsihed_futures) < len(futures):
                    for i, future in enumerate(futures):
                        if future.done() and i not in finsihed_futures:
                            pbar.update(1)
                            finsihed_futures.add(i)
                    sleep(.1)


import os
from sssekai.unity.AssetBundle import load_assetbundle
from UnityPy.enums import ClassIDType
from logging import getLogger
logger = getLogger(__name__)

def main_spineextract(args):
    outdir = args.outdir
    os.makedirs(outdir, exist_ok=True)
    with open(args.infile, "rb") as f:
        env = load_assetbundle(f)
        objects = [pobj.read() for pobj in env.objects]
        binaries = {obj.name: obj for obj in objects if getattr(obj,'type',None) in {ClassIDType.TextAsset}}
        textures = {obj.name: obj for obj in objects if getattr(obj,'type',None) in {ClassIDType.Texture2D}}
        spines = set()
        for name in binaries:
            if name.endswith(".atlas") or name.endswith(".skel"):
                spines.add(".".join(name.split(".")[:-1]))
        for spine in spines:
            logger.info('Extracting %s' % spine)
            atlas = binaries.get(spine + ".atlas", None)
            skel = binaries.get(spine + ".skel", None)
            os.makedirs(os.path.join(outdir, spine), exist_ok=True)
            if atlas:
                logger.info('...has Atlas %s' % spine)
                with open(os.path.join(outdir, spine, spine + ".atlas.txt"), "wb") as f:
                    f.write(atlas.script)
                texfiles = [line.strip() for line in atlas.text.split("\n")]
                texfiles = ['.'.join(line.split('.')[:-1]) for line in texfiles if (line.endswith(".png"))]
                for tex in texfiles:
                    logger.info('...has Texture %s' % tex)
                    textureobj = textures.get(tex, None)
                    if textureobj:
                        textureobj.image.save(os.path.join(outdir, spine, tex + '.png'))
                    else:
                        logger.warning('No texture found for %s' % tex)
            else:
                logger.warning("No atlas found for %s. Consequnetially, no textures cannot be exported either." % spine)

            if skel:
                logger.info('...has Skeleton %s' % spine)
                with open(os.path.join(outdir, spine, spine + ".skel.bytes"), "wb") as f:
                    f.write(skel.script)
            else:
                logger.warning("No skel found for %s" % spine)



from sssekai.unity.AssetBundle import load_assetbundle
from os import path,makedirs
from logging import getLogger
import json
logger = getLogger(__name__)

def main_live2dextract(args):
    with open(args.infile,'rb') as f:
        from UnityPy.enums import ClassIDType
        makedirs(args.outdir,exist_ok=True)
        env = load_assetbundle(f)
        monobehaviors = dict()
        textures = dict()
        animations = dict()
        for obj in env.objects:
            data = obj.read()
            if obj.type in {ClassIDType.MonoBehaviour}:
                monobehaviors[data.name] = data
            if obj.type in {ClassIDType.Texture2D}:
                textures[data.name] = data
            if obj.type in {ClassIDType.AnimationClip}:
                animations[data.name] = data
        modelData = monobehaviors.get('BuildModelData',None)
        if not modelData:
            logger.warning('BuildModelData absent. Not extracting Live2D models!')
        else:
            modelData = modelData.read_typetree()
            # TextAssets are directly extracted
            # Usually there are *.moc3, *.model3, *.physics3; the last two should be renamed to *.*.json
            for obj in env.objects:
                if obj.type == ClassIDType.TextAsset:
                    data = obj.read()
                    out_name : str = data.name
                    if out_name.endswith('.moc3') or out_name.endswith('.model3') or out_name.endswith('.physics3'):
                        if out_name.endswith('.model3') or out_name.endswith('.physics3'):
                            out_name += '.json'
                        with open(path.join(args.outdir, out_name),'wb') as fout:
                            logger.info('Extracting Live2D Asset %s' % out_name)
                            fout.write(data.script)
            # Textures always needs conversion and is placed under specific folders
            for texture in modelData['TextureNames']:
                name = path.basename(texture)
                folder = path.dirname(texture)
                out_folder = path.join(args.outdir, folder)
                makedirs(out_folder,exist_ok=True)
                out_name = path.join(out_folder, name)
                logger.info('Extracting Texture %s' % out_name)
                name_wo_ext = '.'.join(name.split('.')[:-1])
                textures[name_wo_ext].image.save(out_name)
        # Animations are serialized into AnimationClip
        if not args.no_anim:
            from sssekai.unity.constant.SekaiLive2DPathNames import NAMES_CRC_TBL
            from sssekai.fmt.motion3 import unity_animation_clip_to_motion3
            for clipName, clip in animations.items():
                logger.info('Extracting Animation %s' % clipName)
                data = unity_animation_clip_to_motion3(clip, NAMES_CRC_TBL)               
                json.dump(data, open(path.join(args.outdir, clipName+'.motion3.json'),'w'), indent=4, ensure_ascii=False)



import zipfile
import UnityPy
import argparse
import re

from io import BytesIO

HASHREGEX = re.compile(b'[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}')

def dump_axml_stringpool(f : BytesIO):
    read_int = lambda nbytes: int.from_bytes(f.read(nbytes), 'little')
    f.seek(8)
    hdr_type,hdr_size,size = read_int(2),read_int(2),read_int(4)
    n_strings = read_int(4)
    unk = read_int(4)
    flags = read_int(4)
    is_utf8 = flags & (1 << 8)
    string_offset = read_int(4) + 8
    unk = read_int(4)
    offsets = [read_int(4) for _ in range(n_strings)]
    for offset in offsets:
        f.seek(string_offset + offset)
        string_len = read_int(2)
        if is_utf8:
            string = f.read(string_len).decode('utf-8')
        else:
            string = f.read(string_len*2).decode('utf-16-le')
        yield string

def enum_candidates(zip_file, filter):
    return (
        (f,zip_file.open(f),zip_file) for f in zip_file.filelist if filter(f.filename)
    )

def enum_package(zip_file):
    yield zip_file
    for f in zip_file.filelist:        
        if f.filename.lower().endswith('.apk'):
            yield zipfile.ZipFile(zip_file.open(f)) 

def main_apphash(args):
    env = UnityPy.Environment()
    if not args.apk_src or args.fetch:
        from requests import get
        src = BytesIO()
        print('Fetching latest game package from APKPure')
        resp = get('https://d.apkpure.net/b/XAPK/com.sega.pjsekai?version=latest', stream=True)
        size = resp.headers.get('Content-Length',-1)
        for chunck in resp.iter_content(chunk_size=2**10):
            src.write(chunck)
            print('Downloading %d/%s' % (src.tell(), size), end='\r')
        print()
        src.seek(0)
    else:
        src = open(args.apk_src, 'rb')
    with zipfile.ZipFile(src, 'r') as zip_ref:
        manifests = [manifest for package in enum_package(zip_ref) for manifest in enum_candidates(package, lambda fn: fn == 'AndroidManifest.xml')]
        manifest = manifests[0][1]
        manifest_strings = list(dump_axml_stringpool(manifest))
        for i in range(len(manifest_strings)): 
            if 'STAMP_TYPE_' in manifest_strings[i]:
                print('Package Type:', manifest_strings[i])
                print('Version:', manifest_strings[i - 2])
                break
        candidates = [candidate for package in enum_package(zip_ref) for candidate in enum_candidates(
            package, lambda fn: fn.split('/')[-1] in {
            '6350e2ec327334c8a9b7f494f344a761', # PJSK Android
            'c726e51b6fe37463685916a1687158dd'  # PJSK iOS
        })]
        for candidate,stream,_ in candidates:
            env.load_file(stream)
        for obj in env.objects:
            obj = obj.read()
            hashStr = HASHREGEX.finditer(obj.raw_data)
            for m in hashStr:
                print(m.group().decode(), obj.name)


import sys,os

sys.path.insert(0, os.path.dirname(os.path.abspath(__file__)))
import setuptools, sssekai

with open("README.md", "r", encoding='utf-8') as fh:
    long_description = fh.read()

setuptools.setup(
    name="sssekai",
    version=sssekai.__version__,
    author="greats3an",
    author_email="greats3an@gmail.com",
    description="Project SEKAI Asset Utility / PJSK 资源下载 + Live2D, Spine, USM 提取",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/mos9527/sssekai",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    install_requires=["msgpack","pycryptodome","unitypy >= 1.10.14","wannacri", "tqdm", "coloredlogs", "requests"],
    entry_points={"console_scripts": ["sssekai = sssekai.__main__:__main__"]},
    python_requires=">=3.10",
)



import os, re, json
from sssekai.abcache import AbCache, AbCacheConfig, logger
from sssekai.crypto.AssetBundle import SEKAI_AB_MAGIC, decrypt_headaer_inplace
from concurrent.futures import ThreadPoolExecutor
from requests import Session
from tqdm import tqdm

DEFAULT_CACHE_DB_FILE = '~/.sssekai/abcache.db'

class AbCacheDownloader(ThreadPoolExecutor):
    session : Session
    progress : tqdm = None

    def _ensure_progress(self):
        if not self.progress:
            self.progress = tqdm(bar_format="{desc}: {percentage:.1f}%|{bar}| {n_fmt}/{total_fmt} {rate_fmt} {elapsed}<{remaining}", total=0,unit='B', unit_scale=True, unit_divisor=1024,desc="Downloading")
   
    def _download(self, url, fname, length):
        self._ensure_progress()
        RETRIES = 1
        for _ in range(0,RETRIES):
            try:
                resp = self.session.get(url,stream=True)
                resp.raise_for_status()
                os.makedirs(os.path.dirname(fname),exist_ok=True)
                with open(fname, 'wb') as f:
                    magic = next(resp.iter_content(4))
                    self.progress.update(4)
                    if magic == SEKAI_AB_MAGIC:
                        header = next(resp.iter_content(128))            
                        self.progress.update(128)
                        f.write(decrypt_headaer_inplace(bytearray(header)))
                    else:
                        f.write(magic)                
                    for chunk in resp.iter_content(65536):
                        self.progress.update(len(chunk))
                        f.write(chunk)
                    return
            except Exception as e:
                logger.error('While downloading %s : %s. Retrying' % (url,e))
        if _ == RETRIES - 1:
            logger.critical('Did not download %s' % url)
        self._ensure_progress()
    
    def __init__(self, session, **kw) -> None:
        self.session = session        
        super().__init__(**kw)
    
    def __enter__(self):
        return super().__enter__()

    def __exit__(self, exc_type, exc_val, exc_tb):
        return super().__exit__(exc_type, exc_val, exc_tb)

    def add_link(self, url, fname, length):
        self._ensure_progress()
        self.progress.total += length
        return self.submit(self._download, url, fname, length)

def main_abcache(args):    
    cache = AbCache(AbCacheConfig(args.app_version, args.app_platform, args.app_appHash))
    if args.dump_master_data:
        logger.info('Dumping master data to %s', args.dump_master_data)
        cache.update_client_headers()
        for split in tqdm(cache.database.sekai_user_auth_data.suiteMasterSplitPath, desc='Pulling'):
            resp = cache.request_packed('GET', cache.SEKAI_API_ENDPOINT + '/api/' + split)
            data = cache.response_to_dict(resp)  
            path = os.path.join(args.dump_master_data, split + '.json')
            os.makedirs(os.path.dirname(path), exist_ok=True)
            logger.debug('Saving to %s', path)
            with open(path, 'w', encoding='utf-8') as f:
                json.dump(data, f, indent=4, ensure_ascii=False)
            return  

    db_path = os.path.expanduser(args.db)
    os.makedirs(os.path.dirname(db_path), exist_ok=True)
    if not args.no_update:
        assert args.app_version and args.app_appHash, "You need --app-version and --app-appHash to perform a cache index update!"
        with open(db_path, 'wb') as f:
            cache.update()
            cache.save(f)
    else:
        with open(db_path, 'rb') as f:
            cache.load(f)
    logger.info("AbCache: %s" % cache)
    
    if args.download_dir:
        download_dir = os.path.expanduser(args.download_dir)
        bundles = set()
        def _iter_bundles():
            if args.download_filter:
                logger.info('Filtering bundles with regex pattern: %s', args.download_filter)
                pattern = re.compile(args.download_filter)
                for entry in cache.abcache_index.bundles:
                    if pattern.match(entry): yield entry
            else:
                logger.warning('No filter pattern specified. All bundles will be downloaded.')
                for entry in cache.abcache_index.bundles: yield entry
        for bundleName in _iter_bundles(): bundles.add(bundleName)
        basebundles = bundles.copy()
        logger.info('Selected %d bundles to download', len(basebundles))
        if args.download_ensure_deps:
            for bundleName in basebundles:
                cache.get_or_update_dependency_tree_flatten(bundleName, bundles)
            logger.info('Added dependencies:')
            for dep in bundles - basebundles:
                logger.info('   - %s', dep) 
        with AbCacheDownloader(cache, max_workers=args.download_workers) as downloader:
            cache.update_download_headers()
            logger.info('Downloading %d bundles to %s' % (len(bundles), download_dir))
            for bundleName in bundles:
                entry = cache.get_entry_by_bundle_name(bundleName)
                url = cache.get_entry_download_url(entry)
                fname = os.path.join(download_dir, bundleName)
                downloader.add_link(url, fname, entry.fileSize)



from sssekai.unity.AnimationClip import read_animation
from UnityPy.classes import AnimationClip
from logging import getLogger
logger = getLogger(__name__)

# Thanks! https://github.com/Perfare/UnityLive2DExtractor/blob/master/UnityLive2DExtractor/CubismMotion3Converter.cs
def unity_animation_clip_to_motion3(animationClip: AnimationClip, pathTable : dict) -> dict:
    '''Convert Unity AnimationClip to Live2D Motion3 format

    Args:
        animationClip (AnimationClip): animationClip
        pathTable (dict): CRC32 to Live2D path table

    Returns:
        dict: Live2D Motion3 data
    '''
    motion = {
            'Version': 3,
            'Meta': {
                "Name":animationClip.m_Name,
                "Duration":animationClip.m_MuscleClip.m_StopTime,
                "Fps":animationClip.m_SampleRate,
                "Loop":True,
                "AreBeziersRestricted":True,
                "CurveCount":0,
                "UserDataCount":0,
                "TotalPointCount":0,
                "TotalSegmentCount":0,
                "TotalUserDataSize":0
            },
            'Curves': [],
            'UserData': []
        }
    animation = read_animation(animationClip)
    floatCurves = animation.FloatTracks
    motion['Meta']['CurveCount'] = len(floatCurves)
    for track in [track for path in floatCurves.values() for track in path.values()]:
        segments = list()
        segments.append(0)
        segments.append(track.Curve[0].value)
        curveIndex = 1
        while curveIndex < len(track.Curve):
            curve = track.Curve[curveIndex]
            preCurve = track.Curve[curveIndex - 1]
            if (abs(curve.time - preCurve.time - 0.01) < 0.0001): 
                nextCurve = track.Curve[curveIndex + 1]
                if (nextCurve.value == curve.value):
                    segments.append(3) # InverseSteppedSegment
                    segments.append(nextCurve.time)
                    segments.append(nextCurve.value)
                    motion['Meta']['TotalPointCount'] += 1
                    motion['Meta']['TotalSegmentCount'] += 1
                    curveIndex += 1
                    continue
            if (curve.inSlope == float('+inf')):
                segments.append(2) # SteppedSegment
                segments.append(curve.time)
                segments.append(curve.value)
                motion['Meta']['TotalPointCount'] += 1
            elif (preCurve.outSlope == 0 and abs(curve.inSlope) < 0.0001):
                segments.append(0) # LinearSegment
                segments.append(curve.time)
                segments.append(curve.value)
                motion['Meta']['TotalPointCount'] += 1
            else:
                tangentLength = (curve.time - preCurve.time) / 3
                segments.append(1) # BezierSegment
                segments.append(preCurve.time + tangentLength)
                segments.append(preCurve.outSlope * tangentLength + preCurve.value)
                segments.append(curve.time - tangentLength)
                segments.append(curve.value - curve.inSlope * tangentLength)
                segments.append(curve.time)
                segments.append(curve.value)
                motion['Meta']['TotalPointCount'] += 3
            motion['Meta']['TotalSegmentCount'] += 1
            curveIndex+=1
        path = track.Path
        if path in pathTable:
            target, id = pathTable[path].split('/')
            if target == 'Parameters': target = 'Parameter'
            if target == 'Parts': target = 'PartOpacity'
        else:
            logger.warning('Failed to bind path CRC %s to any Live2D path' % path)
            target,id = 'PartOpacity', str(path)
        motion['Curves'].append(
            {
                'Target' : target,
                'Id': id,
                'Segments': segments
            }
        )    
    for event in animationClip.m_Events:
        motion['UserData'].append(
            {
                'time' : event.time,
                'value': event.data
            }
        )
        motion['Meta']['UserDataCount'] += 1
        motion['Meta']['TotalUserDataSize'] += len(event.data)
    return motion
